﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Net;
using System.IO;
using System.Runtime.InteropServices;
using System.Diagnostics;

namespace mp3_15_ctrl {

    public partial class Form1: Form {
        class WinCtrlCmd {
            public byte cmd;
            public byte[] receiveBytes;
        }

#if DEBUG
        private const String LOCAL_DIR = @"C:\UserData\Source\Ele\03_MP3プレーヤ\MP3プレーヤ15(VS1011+PIC18F26K22+ESP-WROOM-02)\WinCtrl\";
#else
        private const String LOCAL_DIR = "";
#endif

        // DIR構成ファイル(Version=15)
        private const String DIRSTRUCTUREFILE = "DIRLST15.DAT";
        private const int RECLEN_DIRLIST = 192;             // レコード長
        private const int DL_DISPLAYNAME_LEN = 32;          // ディレクトリ表示名(SJIS)のバイト数
        private const int MAXCHARS_DIRNAMEU = 30;           // ディレクトリ表示名(UNICODE)の最大文字数
        private const int DL_DISPLAYNAMEU_OFFSET = 32;
        private const int DL_DISPLAYNAMEU_LEN = 62;
        private const int DL_FULLDIR_OFFSET = 94;
        private const int DL_FULLDIR_LEN = 61;
        private const int DL_FULLDIRBYTES_OFFSET = 155;
        private const int DL_FULLDIRBYTES_LEN = 1;
        private const int DL_DIRNAME_OFFSET = 156;
        private const int DL_DIRNAME_LEN = 13;
        private const int DL_RECNO_SUBDIR_START_OFFSET = 169;
        private const int DL_RECNO_SUBDIR_START_LEN = 2;
        private const int DL_NUM_SUBDIR_OFFSET = 171;
        private const int DL_NUM_SUBDIR_LEN = 1;
        private const int DL_RECNO_PARENTDIR_OFFSET = 172;
        private const int DL_RECNO_PARENTDIR_LEN = 2;
        private const int DL_RECNO_GRANDPARENTDIR_OFFSET = 174;
        private const int DL_RECNO_GRANDPARENTDIR_LEN = 2;
        private const int DL_NUM_SNDFILE_OFFSET = 176;
        private const int DL_NUM_SNDFILE_LEN = 1;
        private const int DL_RECNO_SNDFILE_OFFSET = 177;
        private const int DL_RECNO_SNDFILE_LEN = 2;
        private const int DL_FILLER_OFFSET = 179;
        private const int DL_FILLER_LEN = 13;

        // MP3ファイルリスト(Version=15)
        private const string SNDFILELISTFILE = "SNDLST15.DAT";
        private const int RECLEN_MP3LIST = 128;                 // イメージ部分を除くレコード長
        private const int ML_TITLENAME_LEN = 32;
        private const int ML_TITLENAMEU_OFFSET = 32;
        private const int ML_TITLENAMEU_LEN = 74;
        private const int ML_FILENAME_OFFSET = 106;
        private const int ML_FILENAME_LEN = 13;
        private const int ML_TIMESEC_OFFSET = 119;
        private const int ML_TIMESEC_LEN = 2;
        private const int ML_FILESIZE_OFFSET = 121;
        private const int ML_FILESIZE_LEN = 4;
        private const int ML_TAGSIZE_OFFSET = 125;
        private const int ML_TAGSIZE_LEN = 2;
        private const int ML_FILLER_OFFSET = 127;
        private const int ML_FILLER_LEN = 1;

        // MP3_15設定ファイル
        private const string INIFILE = "CONFIG15.INI";

        // マスタファイル作成コマンド
        private const String MKFLMP3 = "MKFLMP3.EXE";

        [DllImport("kernel32.dll", EntryPoint = "GetPrivateProfileStringW", CharSet = CharSet.Unicode, SetLastError = true)]
        static extern uint GetPrivateProfileString(string lpAppName, string lpKeyName, string lpDefault, StringBuilder lpReturnedString, uint nSize, string lpFileName);

        // ホストとの通信コマンドコード
        private const Byte WINCTRL_CMD_STARTUP = 0x01;
        private const Byte WINCTRL_CMD_DATATIMESTAMP = 0x02;
        private const Byte WINCTRL_CMD_SHUTDOWN = 0x03;
        private const Byte WINCTRL_CMD_PLAYDIR = 0x10;
        private const Byte WINCTRL_CMD_PLAYRANDOM = 0x12;
        private const Byte WINCTRL_CMD_STOPPLAY = 0x13;
        private const Byte WINCTRL_CMD_NEXT = 0x14;
        private const Byte WINCTRL_CMD_RESUME = 0x15;
        private const Byte WINCTRL_CMD_VOLUP = 0x16;
        private const Byte WINCTRL_CMD_VOLDOWN = 0x17;
        private const Byte WINCTRL_CMD_FASTFORWARD = 0x18;
        private const Byte WINCTRL_CMD_REWIND = 0x19;
        private const Byte WINCTRL_CMD_CTRLACTIVE = 0x40;
        private const Byte WINCTRL_CMD_WAITING = 0x41;
        private const Byte WINCTRL_CMD_NOWPLAYING = 0x42;
        private const Byte WINCTRL_CMD_CONFIRM = 0x43;
        private const Byte WINCTRL_CMD_MESSAGE = 0x80;

        private const Byte MESSAGE_UNKNOWN_COMMAND = 1;
        private const Byte MESSAGE_RECEIVE_OVERRUN = 2;

        private const int NOWPLAYING_FLAG_START = 1;        // 再生開始
        private const int NOWPLAYING_FLAG_HALFWAY = 2;      // 再生途中
        private const int NOWPLAYING_FLAG_END = 3;          // 再生終了

        private const int MAX_VOLUME = 0xD0;                // ボリューム値の最大値（＝最小音量）
        private const int WINCTRL_CMDLENGTH = 16;           // ソケットを通じて送受信する１単位のデータサイズ(BYTE)
        private const int NO_CTRLACCESS_CONFIRM = (15 * 1000);  // (5 * 1000);  // ミリ秒
        private const int NO_CTRLACCESS_TIMEOUT = (3 * 60 * 1000);  // ミリ秒
        private const int TREEVIEW_REDRAW_TIMEOUT = (10 * 1000);  // ミリ秒
        private const int WHEEL_UPDOWN_TIME = 60;        // ホイールによる時間移動単位(秒)
        private const int LEAVE_VOLUME_TIMEOUT = (5 * 1000);    // txtVolumeがフォーカスを得てから自動的に抜けるまでの時間

        private FileStream SndListReader = null;    // MP3ファイルリスト
        private static String DataTimeStampLocal;   // LOCAL側データファイルのタイムスタンプ
        private String DirFilePath;                 // DIR構成ファイル・ファイルパス
        private String SndFilePath;                 // MP3ファイルリスト・ファイルパス
        private String SdRootDir;                   // メンテナンスモードでのSDメモリのルートディレクトリ
        private bool ConnectedToHost = false;       // 
        private bool NowPlaying = false;            // 再生中のときtrue
        private TreeNode[] AllTreeNode = null;      // ディレクトリレコード番号 対 TreeNode の配列
        private TreeNode PlayingTreeNode = null;    // 再生中のノード
        private int PlayingTrackno = 0;             // 再生中の曲番号
        private Int32 CtrlActiveTickCount;          // HOSTとの接続確認時のTickCount
        private UInt16 ElapsedTime;                 // 再生経過時間（秒）
        private Int32 TreeViewSelectTickCount;      // TreeViewでディレクトリ選択操作をしたときのTickCount
        private int NowPlayingTracknoTmp = -1;      // 再生中の曲番号TMP
        private bool ConfirmHost = false;           // WINCTRL_CMD_CONFIRMの送信状態
        private Int32 ConfirmTickCount;             // WINCTRL_CMD_CONFIRMの送信TickCount
        private Int32 TxtVolumeEnterTickCount;      // txtVolumeのフォーカス取得

        private String HostIp;      // ホストのIPアドレス
        private UInt16 HostPort;     // ホストのポート番号
        private UInt16 LocalPort;    // ローカルのポート番号

        private UdpClient recvUdpClient = null;
        private IPEndPoint RemoteIpEndPoint = null;

        //private bool messageReceived = false;
        //private byte[] receiveBytes = null;
        // 以下、HOSTからのコマンド管理
        private const int RECEIVED_CMD_STACK_SIZE = 10;
        private WinCtrlCmd[] ReceivedCmdStack = new WinCtrlCmd[RECEIVED_CMD_STACK_SIZE];
        private int WinCtrlCmdLast = 0;
        private readonly object ReceivedCmdStackLock = new object();

        private byte CmdSeq = 0;

        // 本PGの動作モード
        enum STATUS { PLAY, MAINTENANCE };      // PLAY=ホスト上のファイルを再生、MAINTENANCE=SDカードのメンテナンス
        STATUS RunningMode;

        ContextMenuStrip ctmTreeViewDir;

        public Form1()
        {
            InitializeComponent();
            txtVolume.MouseWheel += new System.Windows.Forms.MouseEventHandler(this.txtVolume_MouseWheel);
            txtPlayTime.MouseWheel += new System.Windows.Forms.MouseEventHandler(this.txtPlayTime_MouseWheel);
        }

        /// <summary>
        /// PLAYボタン
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnPlay_Click(object sender, EventArgs e)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];

                if (ConnectedToHost = true && NowPlaying == false) {
                    UInt16 recno_dir;
                    UInt16 trackno = 0;
                    //UInt32 startbyteposition = 0;
                    DirInfo dir = (DirInfo)(treeViewDir.SelectedNode.Tag);

                    if (dir != null && dir.num_sndfile > 0) {       // MP3ファイルの存在するディレクトリを選択している
                        recno_dir = dir.recno_curdir;
                        if (listViewMp3.SelectedItems.Count > 0) {
                            trackno = (UInt16)(listViewMp3.SelectedItems[0].Index);
                        } else {
                            trackno = 0;
                        }
                        // MP3ファイルの再生開始秒
                        int psec = GetTxtPlayTime();

                        // 再生開始
                        buf[0] = WINCTRL_CMD_PLAYDIR;
                        buf[1] = (byte)(recno_dir & 0xFF);
                        buf[2] = (byte)((recno_dir >> 8) & 0xFF);
                        buf[3] = (byte)trackno;
                        buf[4] = (byte)(psec & 0xFF);
                        buf[5] = (byte)((psec >> 8) & 0xFF);
                        Array.Clear(buf, 6, WINCTRL_CMDLENGTH - 6);
                        //buf[4] = (byte)(startbyteposition & 0xFF);
                        //buf[5] = (byte)((startbyteposition >> 8) & 0xFF);
                        //buf[6] = (byte)((startbyteposition >> 16) & 0xFF);
                        //buf[7] = (byte)((startbyteposition >> 24) & 0xFF);
                        //Array.Clear(buf, 8, WINCTRL_CMDLENGTH - 8);
                        buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                        buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                        SendCmdByUdp(buf);
                        ElapsedTime = 0;

                        if (PlayingTreeNode != null) {
                            SetTreeNodeProp_Default(PlayingTreeNode);
                        }
                        SetTreeNodeProp_Playing(treeViewDir.SelectedNode);
                        PlayingTreeNode = treeViewDir.SelectedNode;
                        SetListViewItemProp_Playing(listViewMp3.Items[trackno]);
                    }
                }

            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// 再開
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnResume_Click(object sender, EventArgs e)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];
                if (ConnectedToHost == true && NowPlaying == false) {
                    // 前回再生の再開（履歴はHOSTで管理する）
                    buf[0] = WINCTRL_CMD_RESUME;
                    Array.Clear(buf, 1, WINCTRL_CMDLENGTH - 1);
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                }

            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// ランダム再生ボタン
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnRandom_Click(object sender, EventArgs e)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];

                buf[0] = WINCTRL_CMD_PLAYRANDOM;
                for (int i = 1; i < WINCTRL_CMDLENGTH; i++) {
                    buf[i] = 0;
                }
                buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                SendCmdByUdp(buf);
            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// 次へ　ボタン
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnNext_Click(object sender, EventArgs e)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];

                if (NowPlaying == true) {
                    buf[0] = WINCTRL_CMD_NEXT;
                    for (int i = 1; i < WINCTRL_CMDLENGTH; i++) {
                        buf[i] = 0;
                    }
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                }
            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// 再生停止ボタン
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnStop_Click(object sender, EventArgs e)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];

                if (NowPlaying == true) {
                    buf[0] = WINCTRL_CMD_STOPPLAY;
                    for (int i = 1; i < WINCTRL_CMDLENGTH; i++) {
                        buf[i] = 0;
                    }
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                }
            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        private void btnVolUp_Click(object sender, EventArgs e)
        {
            VolUp(4);
        }

        private void VolUp(byte step)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];

                if (NowPlaying == true) {
                    buf[0] = WINCTRL_CMD_VOLUP;
                    buf[1] = step;
                    for (int i = 2; i < WINCTRL_CMDLENGTH; i++) {
                        buf[i] = 0;
                    }
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                }
            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        private void btnVolDown_Click(object sender, EventArgs e)
        {
            VolDown(4);
        }

        private void VolDown(byte step)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];

                if (NowPlaying == true) {
                    buf[0] = WINCTRL_CMD_VOLDOWN;
                    buf[1] = step;
                    for (int i = 2; i < WINCTRL_CMDLENGTH; i++) {
                        buf[i] = 0;
                    }
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                }
            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// 曲内先送り
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnForward_Click(object sender, EventArgs e)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];
                UInt32 skiptime;

                if (NowPlaying == true) {
                    skiptime = 60;      // 秒

                    buf[0] = WINCTRL_CMD_FASTFORWARD;
                    buf[1] = (byte)(skiptime & 0xFF);
                    buf[2] = (byte)((skiptime >> 8) & 0xFF);
                    Array.Clear(buf, 3, WINCTRL_CMDLENGTH - 3);
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                }
            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// 曲内先送り(大)
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnForwardLarge_Click(object sender, EventArgs e)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];
                UInt32 skiptime;

                if (NowPlaying == true) {
                    skiptime = 60 * 10;       // 秒

                    buf[0] = WINCTRL_CMD_FASTFORWARD;
                    buf[1] = (byte)(skiptime & 0xFF);
                    buf[2] = (byte)((skiptime >> 8) & 0xFF);
                    Array.Clear(buf, 3, WINCTRL_CMDLENGTH - 3);
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                }
            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// 曲内巻き戻し
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnRewind_Click(object sender, EventArgs e)
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];
                UInt32 skiptime;

                if (NowPlaying == true) {
                    skiptime = 15;          // 秒

                    buf[0] = WINCTRL_CMD_REWIND;
                    buf[1] = (byte)(skiptime & 0xFF);
                    buf[2] = (byte)((skiptime >> 8) & 0xFF);
                    Array.Clear(buf, 3, WINCTRL_CMDLENGTH - 3);
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                }
            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// チェックサムの計算
        /// </summary>
        /// <param name="buf"></param>
        /// <param name="bytes"></param>
        /// <returns></returns>
        private byte CalcCheckSum(Byte[] buf, int bytes)
        {
            UInt16 cs = 0;
            int i;

            for (i = 0; i < bytes; i++) {
                cs += buf[i];
            }

            return (byte)(cs & 0xFF);
        }

        /// <summary>
        /// コマンドバイトをホストにUDP送信
        /// </summary>
        /// <param name="cmdbyte"></param>
        private void SendCmdByUdp(Byte[] cmdbyte)
        {
            UdpClient connectToHost = null;
            connectToHost = new UdpClient(HostIp, HostPort);
            connectToHost.Send(cmdbyte, WINCTRL_CMDLENGTH);
            connectToHost.Close();
        }


        private void Form1_Load(object sender, EventArgs e)
        {
            //SetRunningMode();
            RunningMode = STATUS.PLAY;

            // コンテキストメニュー作成
            CreateContextMenu();

            ReadIniFile();
            RemoteIpEndPoint = new IPEndPoint(IPAddress.Any, LocalPort /*4211*/);
            recvUdpClient = new UdpClient(RemoteIpEndPoint);
            recvUdpClient.BeginReceive(new AsyncCallback(ReceiveCallback), recvUdpClient);

            SetControlsNoConnect();

        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (ConnectedToHost == true) {
                if (NowPlaying == true) {
                    e.Cancel = true;
                } else {
                    DialogResult res;
                    res = MessageBox.Show("ホストに接続中。ホストも終了？(はい=ホストも終了, いいえ=ホストは終了しない", "",
                        MessageBoxButtons.YesNoCancel);
                    if (res == DialogResult.Yes) {
                        btnConnect_Click(null, null);
                    } else if (res == DialogResult.No) {
                    } else {
                        e.Cancel = true;
                    }
                }
            }
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            try {
                SndListReader.Close();
            }catch (Exception) {
            }
        }

        /// <summary>
        /// INIファイルによる設定取得
        /// </summary>
        private void ReadIniFile()
        {
            StringBuilder sb = new StringBuilder(64);

            uint ret = GetPrivateProfileString("MP3_15", "IP", "", sb, Convert.ToUInt32(sb.Capacity), LOCAL_DIR + INIFILE);
            if (0 < ret) {
                HostIp = sb.ToString();
            }

            ret = GetPrivateProfileString("MP3_15", "PORT", "0", sb, Convert.ToUInt32(sb.Capacity), LOCAL_DIR + INIFILE);
            if (0 < ret) {
                HostPort = UInt16.Parse(sb.ToString());
            }

            ret = GetPrivateProfileString("MP3_15_CTRL", "PORT", "0", sb, Convert.ToUInt32(sb.Capacity), LOCAL_DIR + INIFILE);
            if (0 < ret) {
                LocalPort = UInt16.Parse(sb.ToString());
            }
        }

        /// <summary>
        /// ディレクトリ構成ファイルのタイムスタンプを取得
        /// </summary>
        /// <param name="dirfile">ディレクトリ構成ファイル</param>
        /// <returns></returns>
        private String GetDataTimeStamp(String dirfile)
        {
            FileStream reader = null;
            byte[] RecDirList = new byte[RECLEN_DIRLIST];
            Decoder decoder = Encoding.Unicode.GetDecoder();
            String timestamp;

            try {
                reader = new FileStream(dirfile, FileMode.Open, FileAccess.Read);
            } catch (FileNotFoundException ex) {
                MessageBox.Show("ディレクトリ構成ファイルが存在しません");
                return null;
            }
            try {
                // ヘッダレコード
                if (reader.Read(RecDirList, 0, RECLEN_DIRLIST) != RECLEN_DIRLIST) {
                    throw new Exception("ディレクトリ構成ファイルが不正です");
                }
                timestamp = Encoding.UTF8.GetString(RecDirList, 4, 14);

            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
                return null;
            } finally {
                reader.Close();
            }

            return timestamp;
        }

    /// <summary>
    /// DIR構成ファイルを読み込んで TreeViewに表示
    /// </summary>
    private bool LoadDirStructFile()
        {
            FileStream reader = null;
            byte[] RecDirList = new byte[RECLEN_DIRLIST];
            Decoder decoder = Encoding.Unicode.GetDecoder();

            // ディレクトリ構成ファイルオープン
            try {
                //reader = new FileStream(DIRSTRUCTUREFILE, FileMode.Open, FileAccess.Read);
                reader = new FileStream(DirFilePath, FileMode.Open, FileAccess.Read);
            } catch (FileNotFoundException ex) {
                MessageBox.Show("ディレクトリ構成ファイルが存在しません");
                return false;
            }
            try {
                // ヘッダレコード
                if (reader.Read(RecDirList, 0, RECLEN_DIRLIST) != RECLEN_DIRLIST) {
                    throw new Exception("ディレクトリ構成ファイルが不正です");
                }
                UInt16 numdirs = (UInt16)(RecDirList[0] | ((UInt16)RecDirList[1] << 8));    // ディレクトリ総数(明細レコード総数)
                AllTreeNode = new TreeNode[numdirs + 1];
                DataTimeStampLocal = Encoding.UTF8.GetString(RecDirList, 4, 14);

                // ルートレコード
                if (reader.Read(RecDirList, 0, RECLEN_DIRLIST) != RECLEN_DIRLIST) {
                    throw new Exception("ディレクトリ構成ファイルが不正です");
                }
                DirInfo root = GetDirInfo(RecDirList, 1, decoder);
                reader.Seek((root.recno_subdir_start) * RECLEN_DIRLIST, SeekOrigin.Begin);
                TreeNode[] nodes = new TreeNode[root.num_subdir];
                DirInfo[] infos = new DirInfo[root.num_subdir];

                // ルート直下のサブディレクトリ
                for (int sd = 0; sd < root.num_subdir; sd++) {
                    if (reader.Read(RecDirList, 0, RECLEN_DIRLIST) != RECLEN_DIRLIST) {
                        throw new Exception("ディレクトリ構成ファイルが不正です");
                    }
                    DirInfo subdir = GetDirInfo(RecDirList, (UInt16)(2 + sd), decoder);
                    TreeNode dirnode = new TreeNode(subdir.dirname_w /*, icono, icono*/);
                    dirnode.Tag = subdir;
                    treeViewDir.Nodes.Add(dirnode);
                    nodes[sd] = dirnode;
                    infos[sd] = subdir;
                    AllTreeNode[2 + sd] = dirnode;
                    if (RunningMode == STATUS.MAINTENANCE) {
                        dirnode.ContextMenuStrip = ctmTreeViewDir;
                    }
                }

                for (int sd = 0; sd < root.num_subdir; sd++) {
                    AddTreeNode(reader, RecDirList, decoder, nodes[sd], infos[sd]);
                }

            } catch (Exception ex) {
                MessageBox.Show(ex.Message);
                return false;
            } finally {
                reader.Close();
            }

            return true;
        }

        /// <summary>
        /// DIR構成を表示するTreeViewにノードを追加する
        /// </summary>
        /// <param name="reader"></param>
        /// <param name="RecDirList"></param>
        /// <param name="decoder"></param>
        /// <param name="dirNode"></param>
        /// <param name="dirinfo"></param>
        void AddTreeNode(FileStream reader, byte[] RecDirList, Decoder decoder, TreeNode dirNode, DirInfo dirinfo)
        {
            reader.Seek(dirinfo.recno_subdir_start * RECLEN_DIRLIST, SeekOrigin.Begin);
            TreeNode[] nodes = new TreeNode[dirinfo.num_subdir];
            DirInfo[] infos = new DirInfo[dirinfo.num_subdir];

            for (int sd = 0; sd < dirinfo.num_subdir; sd++) {
                if (reader.Read(RecDirList, 0, RECLEN_DIRLIST) != RECLEN_DIRLIST) {
                    throw new Exception("ディレクトリ構成ファイルが不正です");
                }
                DirInfo subdirinfo = GetDirInfo(RecDirList, (UInt16)(dirinfo.recno_subdir_start + sd), decoder);
                TreeNode subdirnode = new TreeNode(subdirinfo.dirname_w /*, icono, icono*/);
                subdirnode.Tag = subdirinfo;
                dirNode.Nodes.Add(subdirnode);
                nodes[sd] = subdirnode;
                infos[sd] = subdirinfo;
                AllTreeNode[dirinfo.recno_subdir_start + sd] = subdirnode;
                if (RunningMode == STATUS.MAINTENANCE) {
                    subdirnode.ContextMenuStrip = ctmTreeViewDir;
                }
            }

            for (int sd = 0; sd < dirinfo.num_subdir; sd++) {
                AddTreeNode(reader, RecDirList, decoder, nodes[sd], infos[sd]);
            }
        }

        /// <summary>
        /// ディレクトリ構成ファイルの明細レコードからDirInfoを取得
        /// </summary>
        /// <param name="RecDirList">明細レコード　バイナリデータ</param>
        /// <param name="recno">明細レコードのレコード番号</param>
        /// <param name="decoder">UTF16LEデコーダ</param>
        /// <returns></returns>
        private DirInfo GetDirInfo(byte[] RecDirList, UInt16 recno, Decoder decoder)
        {
            DirInfo dir = new DirInfo();
            SByte[] bytes = new SByte[RECLEN_DIRLIST];

            int cn = decoder.GetCharCount(RecDirList, DL_DISPLAYNAMEU_OFFSET, DL_DISPLAYNAMEU_LEN, true);
            Char[] chars = new char[cn];
            decoder.GetChars(RecDirList, DL_DISPLAYNAMEU_OFFSET, DL_DISPLAYNAMEU_LEN, chars, 0, true);
            dir.dirname_w = new String(chars);
            // ↑Decoder.convert()ではどうか？

            if (RunningMode == STATUS.MAINTENANCE) {
                // ディレクトリフルパスが必要になるのはメンテナンスモードだけ
                dir.fulldir = new byte[DL_FULLDIR_LEN];
                Array.Copy(RecDirList, DL_FULLDIR_OFFSET, dir.fulldir, 0, DL_FULLDIR_LEN);
                dir.fulldirbytes = RecDirList[DL_FULLDIRBYTES_OFFSET];
            } else {
                dir.fulldirbytes = 0;
            }
            dir.recno_curdir = recno;
            dir.recno_subdir_start = (UInt16)(RecDirList[DL_RECNO_SUBDIR_START_OFFSET] + ((UInt16)RecDirList[DL_RECNO_SUBDIR_START_OFFSET + 1] << 8));
            dir.num_subdir = RecDirList[DL_NUM_SUBDIR_OFFSET];
            dir.num_sndfile = RecDirList[DL_NUM_SNDFILE_OFFSET];
            dir.recno_sndfile = (UInt16)(RecDirList[DL_RECNO_SNDFILE_OFFSET] + ((UInt16)RecDirList[DL_RECNO_SNDFILE_OFFSET + 1] << 8));

            return dir;
        }

        /// <summary>
        /// 動作モードの設定
        /// </summary>
        private void SetRunningMode()
        {
            String curdrive = System.IO.Directory.GetCurrentDirectory().Substring(0, 1);    // カレントドライブ

            System.IO.DriveInfo drive = new System.IO.DriveInfo(curdrive);
            if (drive.DriveType == DriveType.Removable) {       // SDメモリから起動時はメンテナンスモード
                RunningMode = STATUS.MAINTENANCE;
            } else {    // PLAYモード
                RunningMode = STATUS.PLAY;
            }
        }

        /// <summary>
        /// コンテキストメニュー作成
        /// </summary>
        private void CreateContextMenu()
        {
            ToolStripMenuItem menuItem;

            menuItem = new ToolStripMenuItem();
            menuItem.Text = "エクスプローラ";
            menuItem.Click += new EventHandler(TreeViewDirAddDir_Click);
            ctmTreeViewDir = new ContextMenuStrip();
            ctmTreeViewDir.Items.Add(menuItem);
        }

        /// <summary>
        /// ディレクトリ構成表示TreeViewにディレクトリ追加
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void TreeViewDirAddDir_Click(object sender, EventArgs e)
        {
            Console.WriteLine("TreeViewDirAddDir_Click:" + sender.GetType().Name);
            TreeNode tnode = treeViewDir.SelectedNode;
            if (tnode != null && tnode.Tag != null) {
                DirInfo dirinfo = (DirInfo)tnode.Tag;
                Console.WriteLine(dirinfo.dirname_w);

                Decoder dec = Encoding.Default.GetDecoder();
                int cn = dec.GetCharCount(dirinfo.fulldir, 0, dirinfo.fulldirbytes, true);
                Char[] chars = new char[cn];
                dec.GetChars(dirinfo.fulldir, 0, dirinfo.fulldirbytes, chars, 0, true);
                String targetdir = new String(chars);
                Console.WriteLine(targetdir);

                // エクスプローラ起動
                if (Directory.Exists(SdRootDir + targetdir) == true) {
                    System.Diagnostics.Process.Start(SdRootDir + targetdir);
                } else {
                    System.Diagnostics.Process.Start(SdRootDir);
                }

                //// フォルダ選択
                //FolderBrowserDialog fbDialog = new FolderBrowserDialog();
                //if (fbDialog.ShowDialog() == DialogResult.OK) {
                //    DirectoryInfo di = new DirectoryInfo(fbDialog.SelectedPath);
                //    Console.WriteLine(fbDialog.SelectedPath + " " + di.Name);

                //    String newdir = targetdir + @"\" + di.Name;
                //    if (Directory.Exists(targetdir) == true) {
                //        DirectoryInfo newdi = Directory.CreateDirectory(newdir);

                //        FileInfo[] files = di.GetFiles();
                //        foreach (FileInfo fi in files){
                //            fi.CopyTo(newdir + @"\" + fi.Name);
                //        }

                //        TreeNode dirnode = new TreeNode(di.Name);
                //        tnode.Nodes.Add(dirnode);
                //    } else {
                //        MessageBox.Show("ディレクトリが存在しません:" + targetdir);
                //    }
                //}
                //fbDialog.Dispose();
            }
        }

        private void ReceiveCallback(IAsyncResult ar)
        {
            UdpClient u = (UdpClient)(ar.AsyncState);
            IPEndPoint e = null;

            //Console.WriteLine("Received");

            WinCtrlCmd cmd = new WinCtrlCmd();
            cmd.receiveBytes = u.EndReceive(ar, ref e);
            cmd.cmd = cmd.receiveBytes[0];
            if (WinCtrlCmdLast < RECEIVED_CMD_STACK_SIZE) {
                lock (ReceivedCmdStackLock) {
                    ReceivedCmdStack[WinCtrlCmdLast++] = cmd;
                }
            }
            //receiveBytes = u.EndReceive(ar, ref e);

            recvUdpClient.BeginReceive(new AsyncCallback(ReceiveCallback), recvUdpClient);
            //messageReceived = true;
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            bool elapsedtime_received = false;
            String msg_work = "";

            txtRcvMsg.Text = "";    // １秒おきに消す

            lock (ReceivedCmdStackLock) {
                for (int i = 0; i < WinCtrlCmdLast; i++) {
                    WinCtrlCmd cmd = ReceivedCmdStack[i];

                    Console.WriteLine(String.Format("Rcv={0:X}, seq={1:X}", cmd.cmd, cmd.receiveBytes[14]));
                    if (cmd.cmd == WINCTRL_CMD_DATATIMESTAMP) {
                        String timestamp = Encoding.UTF8.GetString(cmd.receiveBytes, 1, 14);
                        Console.WriteLine("DataTimeStampHost=" + timestamp);
                        DirFilePath = LOCAL_DIR + timestamp + @"\" + DIRSTRUCTUREFILE;
                        SndFilePath = LOCAL_DIR + timestamp + @"\" + SNDFILELISTFILE;
                        // ディレクトリ構成取得＆表示
                        treeViewDir.Nodes.Clear();
                        if (LoadDirStructFile() == false) {
                            //if (String.Equals(DataTimeStampHost, DataTimeStampLocal) == false) {
                            // ローカルデータファイルのタイムスタンプと比較
                            //MessageBox.Show("データファイルのタイムスタンプが一致しません");
                            //msg_work += "データファイルのタイムスタンプが一致しません; ";
                            MessageBox.Show("データファイルなし：" + timestamp);
                            msg_work += "データファイルなし; ";
                        } else {
                            SndListReader = new FileStream(SndFilePath, FileMode.Open, FileAccess.Read);
                            ConnectedToHost = true;
                            SetControlsConnectToHost();
                            btnSdMaintenance.Enabled = false;
                            NotifyActive();
                            CtrlActiveTickCount = Environment.TickCount;
                        }
                    } else if (cmd.cmd == WINCTRL_CMD_WAITING) {
                        NotifyActive();
                        CtrlActiveTickCount = Environment.TickCount;
                        SetControlsAfterStop();
                        NowPlaying = false;
                        msg_work += "WAITING; ";
                        Console.WriteLine("  Waiting SeqNo=" + String.Format("{0:X}", cmd.receiveBytes[15]));
                    } else if (cmd.cmd == WINCTRL_CMD_NOWPLAYING) {
                        UInt16 dirno;
                        UInt16 trackno;
                        UInt16 volume;
                        UInt16 option_flag;

                        NowPlaying = true;
                        SetControlsNowPlaying();

                        dirno = (UInt16)(cmd.receiveBytes[1] | ((UInt16)cmd.receiveBytes[2]) << 8);
                        trackno = cmd.receiveBytes[3];
                        ElapsedTime = (UInt16)(cmd.receiveBytes[8] + ((UInt16)cmd.receiveBytes[9] << 8));
                        volume = (UInt16)(cmd.receiveBytes[10] + ((UInt16)cmd.receiveBytes[11] << 8));
                        option_flag = cmd.receiveBytes[12];
                        elapsedtime_received = true;

                        DirInfo dir = (treeViewDir.SelectedNode != null ? (DirInfo)(treeViewDir.SelectedNode.Tag) : null);
                        if (dir != null && dir.recno_curdir == dirno) {
                            if (trackno < listViewMp3.Items.Count) {
                                if (trackno != PlayingTrackno && PlayingTrackno >= 0 && PlayingTrackno < listViewMp3.Items.Count) {
                                    SetListViewItemProp_Default(listViewMp3.Items[PlayingTrackno]);
                                }
                                //if (option_flag == NOWPLAYING_FLAG_END) {           // 
                                //    //listViewMp3.Items[trackno].Selected = false;
                                //    SetListViewItemProp_Default(listViewMp3.Items[trackno]);
                                //} else {
                                //    //listViewMp3.Items[trackno].Selected = true;
                                //    SetListViewItemProp_Playing(listViewMp3.Items[trackno]);
                                //}
                                SetListViewItemProp_Playing(listViewMp3.Items[trackno]);
                                PlayingTrackno = trackno;
                            }
                        } else {    // TreeViewの選択ノードは再生中のディレクトリでない
                            if ((Environment.TickCount - TreeViewSelectTickCount) > TREEVIEW_REDRAW_TIMEOUT && dirno >= 0 && dirno < AllTreeNode.Length && AllTreeNode[dirno] != null) {
                                NowPlayingTracknoTmp = trackno;         // SelectedNodeを設定すると（↓）AfterSelectイベントが発生する。その際にlistViewMp3の選択も
                                treeViewDir.SelectedNode = AllTreeNode[dirno];      // 再生中のディレクトリを選択
                                if (AllTreeNode[dirno] != PlayingTreeNode) {
                                    if (PlayingTreeNode != null) {
                                        SetTreeNodeProp_Default(PlayingTreeNode);
                                    }
                                    SetTreeNodeProp_Playing(treeViewDir.SelectedNode);
                                    PlayingTreeNode = AllTreeNode[dirno];
                                }
                            }
                        }
                        txtVolume.Text = (MAX_VOLUME - volume).ToString();
                        NotifyActive();
                        CtrlActiveTickCount = Environment.TickCount;
                        msg_work += "NOW PLAYING; ";
                    } else if (cmd.cmd == WINCTRL_CMD_SHUTDOWN) {
                        ConnectedToHost = false;
                        NowPlaying = false;
                        SetControlsNoConnect();
                        msg_work += "SHUTDOWN from HOST; ";
                        timer1.Enabled = false;
                    } else if (cmd.cmd == WINCTRL_CMD_CONFIRM) {            // HOSTからの返信
                        CtrlActiveTickCount = Environment.TickCount;
                        ConfirmHost = false;
                        msg_work += "CONFIRM from HOST; ";
                    } else if (cmd.cmd == WINCTRL_CMD_MESSAGE) {
                        if (cmd.receiveBytes[1] == MESSAGE_UNKNOWN_COMMAND) {
                            Console.WriteLine("Unknown Command");
                        } else if (cmd.receiveBytes[1] == MESSAGE_RECEIVE_OVERRUN) {
                            Console.WriteLine("Receive Overrun Error");
                        }
                    }
                    cmd.receiveBytes = null;
                    ReceivedCmdStack[i] = null;

                    //recvUdpClient.BeginReceive(new AsyncCallback(ReceiveCallback), recvUdpClient);
                }
                WinCtrlCmdLast = 0;
            }

            if (elapsedtime_received == false) {    // NOW_PLAYINGで再生進捗時間が送られてないときは
                ElapsedTime++;                      // コントローラで時間をカウント
            }

            if (NowPlaying == true) {
                txtPlayTime.Text = String.Format("{0:D}:{1:00}", ElapsedTime / 60, ElapsedTime % 60);
            }

            if (ConnectedToHost == true) {
                if (ConfirmHost == false && (Environment.TickCount - CtrlActiveTickCount) > NO_CTRLACCESS_CONFIRM) {
                    // WaitingまたはNowPlayingがしばらく無い。
                    ConfirmHostActivty();
                    msg_work += "CONFIRM to HOST; ";
                    ConfirmHost = true;
                    ConfirmTickCount = Environment.TickCount;
                } else if (ConfirmHost == true && (Environment.TickCount - ConfirmTickCount) > NO_CTRLACCESS_CONFIRM) {
                    ConfirmHostActivty();
                    msg_work += "CONFIRM to HOST; ";
                    ConfirmTickCount = Environment.TickCount;
                }
            }
            if ((Environment.TickCount - CtrlActiveTickCount) > NO_CTRLACCESS_TIMEOUT) {      // HOSTからの通信が所定時間以上ない
                NowPlaying = false;
                ConnectedToHost = false;
                SetControlsNoConnect();
            }
            if (txtVolume.Focused == true && (Environment.TickCount - TxtVolumeEnterTickCount) > LEAVE_VOLUME_TIMEOUT) {    // txtVolumeにフォーカスを残さない
                listViewMp3.Focus();
            }

            txtRcvMsg.Text = msg_work;
            //Console.WriteLine("Timer");
        }

        /// <summary>
        /// 再生中のTreeNodeの表示属性設定
        /// </summary>
        /// <param name="tn"></param>
        private void SetTreeNodeProp_Playing(TreeNode tn)
        {
            tn.ForeColor = Color.White;
            tn.BackColor = Color.Green;
        }

        /// <summary>
        /// DefaultのTreeNode表示属性
        /// </summary>
        /// <param name="tn"></param>
        private void SetTreeNodeProp_Default(TreeNode tn)
        {
            tn.ForeColor = Color.Black;
            tn.BackColor = Color.Empty;
        }

        /// <summary>
        /// 再生中のListViewの表示属性設定
        /// </summary>
        /// <param name="item"></param>
        private void SetListViewItemProp_Playing(ListViewItem item)
        {
            item.ForeColor = Color.White;
            item.BackColor = Color.Green;
        }

        /// <summary>
        /// DefaultのTreeNode表示属性
        /// </summary>
        /// <param name="item"></param>
        private void SetListViewItemProp_Default(ListViewItem item)
        {
            item.ForeColor = Color.Black;
            item.BackColor = Color.Empty;
        }

        /// <summary>
        /// 実行中であることを通知
        /// </summary>
        private void NotifyActive()
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];

                //if (NowPlaying == true) {
                buf[0] = WINCTRL_CMD_CTRLACTIVE;
                Array.Clear(buf, 1, WINCTRL_CMDLENGTH - 1);
                buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                SendCmdByUdp(buf);
                Console.WriteLine("send CtrlActive " + String.Format("{0:x}", buf[WINCTRL_CMDLENGTH - 2]));
                //}
            } catch (Exception ex) {
                //MessageBox.Show(ex.Message);
            }

        }


        /// <summary>
        /// HOSTの動作確認。
        /// </summary>
        private void ConfirmHostActivty()
        {
            try {
                byte[] buf = new byte[WINCTRL_CMDLENGTH];

                //if (NowPlaying == true) {
                buf[0] = WINCTRL_CMD_CONFIRM;
                Array.Clear(buf, 1, WINCTRL_CMDLENGTH - 1);
                buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                SendCmdByUdp(buf);
                Console.WriteLine("send Confirm " + String.Format("{0:x}", buf[WINCTRL_CMDLENGTH - 2]));
                //}
            } catch (Exception ex) {
                //MessageBox.Show(ex.Message);
            }

        }

        /// <summary>
        /// ディレクトリ選択時
        /// 手で選択したときだけでなく、プログラムから選択したときも発生する
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void treeViewDir_AfterSelect(object sender, TreeViewEventArgs e)
        {
            TreeNode dirNode = e.Node;
            bool btnPlayEnabled = false;

            if (dirNode != null) {
                DirInfo dir = (DirInfo)dirNode.Tag;
                if (dir != null) {
                    listViewMp3.Items.Clear();
                    TreeViewSelectTickCount = Environment.TickCount;

                    // アルバムの曲目を表示
                    Decoder d = Encoding.Unicode.GetDecoder();
                    byte[] RecMp3List = new byte[RECLEN_MP3LIST];
                    SndListReader.Seek(dir.recno_sndfile * RECLEN_MP3LIST, SeekOrigin.Begin);
                    for (int i = 0; i < dir.num_sndfile; i++) {
                        if (SndListReader.Read(RecMp3List, 0, RECLEN_MP3LIST) != RECLEN_MP3LIST) {
                            break;
                        }
                        // タイトル
                        int cn = d.GetCharCount(RecMp3List, ML_TITLENAMEU_OFFSET, ML_TITLENAMEU_LEN);
                        Char[] chars = new char[cn];
                        d.GetChars(RecMp3List, ML_TITLENAMEU_OFFSET, ML_TITLENAMEU_LEN, chars, 0);
                        // 長さ
                        uint time_sec = RecMp3List[ML_TIMESEC_OFFSET] | ((uint)RecMp3List[ML_TIMESEC_OFFSET + 1] << 8);
                        String time_ms = String.Format("{0:D}:{1:00}", time_sec / 60, time_sec % 60);
                        // ITEM
                        ListViewItem item = new ListViewItem(new String(chars));
                        item.SubItems.Add(time_ms);
                        listViewMp3.Items.Add(item);
                    }
                    if (dir.num_sndfile > 0) {
                        listViewMp3.Focus();
                        if (NowPlayingTracknoTmp >= 0 && NowPlayingTracknoTmp < dir.num_sndfile) {
                            listViewMp3.Items[NowPlayingTracknoTmp].Selected = true;
                            NowPlayingTracknoTmp = -1;
                        } else {
                            listViewMp3.Items[0].Selected = true;
                        }
                        btnPlayEnabled = true;
                    } else {
                    }
                } else {
                    listViewMp3.Items.Clear();
                }
            } else {
                listViewMp3.Items.Clear();
            }

            if (RunningMode == STATUS.PLAY && NowPlaying == false) {
                btnPlay.Enabled = btnPlayEnabled;
                txtPlayTime.Text = "";
            }

        }

        /// <summary>
        /// 接続・切断
        /// メンテナンスモードのときはマスタファイル作成(MKFLMP3.EXE実行)
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnConnect_Click(object sender, EventArgs e)
        {
            if (RunningMode == STATUS.PLAY) {
                btnSdMaintenance.Enabled = false;
                if (ConnectedToHost == false) {
                    timer1.Enabled = true;
                    byte[] buf = new byte[WINCTRL_CMDLENGTH];
                    buf[0] = WINCTRL_CMD_STARTUP;
                    buf[1] = (byte)(LocalPort & 0xFF);
                    buf[2] = (byte)((LocalPort >> 8) & 0xFF);
                    for (int i = 3; i < WINCTRL_CMDLENGTH; i++) {
                        buf[i] = 0x00;
                    }
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                    // タイムスタンプのチェックを経てから接続状態へ移行する
                } else {
                    byte[] buf = new byte[WINCTRL_CMDLENGTH];
                    buf[0] = WINCTRL_CMD_SHUTDOWN;
                    for (int i = 1; i < WINCTRL_CMDLENGTH; i++) {
                        buf[i] = 0x00;
                    }
                    buf[WINCTRL_CMDLENGTH - 2] = CmdSeq++;
                    buf[WINCTRL_CMDLENGTH - 1] = CalcCheckSum(buf, WINCTRL_CMDLENGTH - 1);
                    SendCmdByUdp(buf);
                    //// HOSTからのレスポンスなしで切断状態へ移行する
                    //ConnectedToHost = false;
                    //SetControlsNoConnect();
                }
            } else if (RunningMode == STATUS.MAINTENANCE) {
                CreateMasterFiles();
            }
        }

        /// <summary>
        /// txtVolumeのフォーカス取得
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void txtVolume_Enter(object sender, EventArgs e)
        {
            TxtVolumeEnterTickCount = Environment.TickCount;
        }

        /// <summary>
        /// マウスホイールで音量調整
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void txtVolume_MouseWheel(object sender, MouseEventArgs e)
        {
            if (e.Delta > 0) {
                VolUp(1);
                //btnVolUp_Click(sender, e);
            } else {
                VolDown(1);
                //btnVolDown_Click(sender, e);
            }
        }

        /// <summary>
        /// 演奏経過時間テキストボックスtxtPlayTimeの値を秒で取得
        /// </summary>
        /// <returns></returns>
        private int GetTxtPlayTime()
        {
            int psec;
            String ptime = txtPlayTime.Text;
            String[] h_m = ptime.Split(new char[] { ':' });
            if (h_m.Length == 2) {
                try {
                    psec = int.Parse(h_m[0]) * 60 + (int)decimal.Parse(h_m[1]);
                } catch (Exception ex) {
                    psec = 0;
                }
            } else {
                psec = 0;
            }
            return psec;
        }

        /// <summary>
        /// 再生開始時間を調整
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void txtPlayTime_MouseWheel(object sender, MouseEventArgs e)
        {
            int psec;

            if (NowPlaying == false) {
                psec = GetTxtPlayTime();
                if (e.Delta > 0) {      // 進める
                    psec += WHEEL_UPDOWN_TIME;
                } else {
                    if (psec >= WHEEL_UPDOWN_TIME) {
                        psec -= WHEEL_UPDOWN_TIME;
                    } else {
                        psec = 0;
                    }
                }
                txtPlayTime.Text = String.Format("{0:D}:{1:00}", psec / 60, psec % 60);
            }
        }

        /// <summary>
        /// コントロールの状態設定：HOSTに接続無し
        /// </summary>
        private void SetControlsNoConnect()
        {
            btnConnect.Text = "接続";
            btnPlay.Enabled = false;
            btnResume.Enabled = false;
            btnNext.Enabled = false;
            btnStop.Enabled = false;
            btnRandom.Enabled = false;
            btnVolUp.Enabled = false;
            btnVolDown.Enabled = false;
            btnForward.Enabled = false;
            btnForwardLarge.Enabled = false;
            btnRewind.Enabled = false;
        }

        /// <summary>
        /// コントロールの状態設定：HOSTに接続後
        /// </summary>
        private void SetControlsConnectToHost()
        {
            btnConnect.Text = "切断";
            //btnPlay.Enabled = true;   // treeViewDir_AfterSelect()で設定する
            btnResume.Enabled = true;
            btnNext.Enabled = false;
            btnStop.Enabled = false;
            btnRandom.Enabled = true;
            btnVolUp.Enabled = false;
            btnVolDown.Enabled = false;
            btnForward.Enabled = false;
            btnForwardLarge.Enabled = false;
            btnRewind.Enabled = false;
        }

        /// <summary>
        /// コントロールの状態設定：再生→停止における設定。ホストはコントローラからのコマンド待ち。
        /// </summary>
        private void SetControlsAfterStop()
        {
            btnPlay.Enabled = true;     // SetControlsConnectToHost()とは異なりbtnPlayも設定する。
            btnResume.Enabled = true;
            btnNext.Enabled = false;
            btnStop.Enabled = false;
            btnRandom.Enabled = true;
            btnVolUp.Enabled = false;
            btnVolDown.Enabled = false;
            btnForward.Enabled = false;
            btnForwardLarge.Enabled = false;
            btnRewind.Enabled = false;
        }

        /// <summary>
        /// コントロールの状態設定：再生中
        /// </summary>
        private void SetControlsNowPlaying()
        {
            btnPlay.Enabled = false;
            btnResume.Enabled = false;
            btnNext.Enabled = true;
            btnStop.Enabled = true;
            btnRandom.Enabled = true;
            btnVolUp.Enabled = true;
            btnVolDown.Enabled = true;
            btnForward.Enabled = true;
            btnForwardLarge.Enabled = true;
            btnRewind.Enabled = true;
        }

        /// <summary>
        /// マスタデータファイル作成処理
        /// MKFLMP3.exe起動
        /// </summary>
        private void CreateMasterFiles()
        {
            try {
                try {
                    SndListReader.Close();
                } catch (Exception) {
                }

                // SDメモリ上でマスタファイル作成
                Process proc = new System.Diagnostics.Process();
                proc.StartInfo.FileName = MKFLMP3;               // @"MKFLMP3.EXE";
                proc.StartInfo.WorkingDirectory = SdRootDir;
                proc.Start();
                proc.WaitForExit();

                // ローカルにマスタファイルをコピー
                String timestamp = GetDataTimeStamp(DirFilePath);
                String localdir = LOCAL_DIR + timestamp;
                Directory.CreateDirectory(localdir);
                string[] datList = Directory.GetFiles(SdRootDir, "*.DAT");
                foreach (string f in datList) {
                    string fName = f.Substring(SdRootDir.Length);
                    File.Copy(f, Path.Combine(localdir, fName), true);
                }

                // TreeViewの再作成
                treeViewDir.Nodes.Clear();
                //SndListReader = new FileStream(SNDFILELISTFILE, FileMode.Open, FileAccess.Read);
                SndListReader = new FileStream(SndFilePath, FileMode.Open, FileAccess.Read);
                LoadDirStructFile();
            } finally {

            }
        }

        /// <summary>
        /// SDメンテナンスボタン
        /// メンテナンスモードに移行する。
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void btnSdMaintenance_Click(object sender, EventArgs e)
        {
            // SDドライブを探す
            // ルートにINIFILEが存在するドライブ
            bool res = false;
            DriveInfo[] allDrives = DriveInfo.GetDrives();

            foreach (DriveInfo d in allDrives) {
                if (d.DriveType == DriveType.Removable) {
                    String filepath = d.Name + INIFILE;
                    if (File.Exists(filepath) == true) {
                        SdRootDir = d.Name;
                        res = true;
                        break;
                    }
                }
            }

            if (res == false) {
                MessageBox.Show("SDドライブが存在しません。SDドライブのルートには " + INIFILE + " が存在すること。");
                System.Diagnostics.Process.Start(LOCAL_DIR);
                return;
            }
            File.Copy(LOCAL_DIR + MKFLMP3, SdRootDir + MKFLMP3, true);

            DirFilePath = SdRootDir + DIRSTRUCTUREFILE;
            SndFilePath = SdRootDir + SNDFILELISTFILE;
            RunningMode = STATUS.MAINTENANCE;
            btnConnect.Text = "作成";

            if (File.Exists(DirFilePath) == true && File.Exists(SndFilePath) == true) {
                treeViewDir.Nodes.Clear();
                try {
                    SndListReader.Close();
                } catch (Exception) {
                }
                if (LoadDirStructFile() == true) {
                    SndListReader = new FileStream(SndFilePath, FileMode.Open, FileAccess.Read);
                }
            }
            MessageBox.Show("エクスプローラでSDドライブにファイルをコピーし作成ボタンを押す。\n\n SDドライブを確認：" + SdRootDir + "\n\n ランダム再生対象のディレクトリには MP3_15_RANDOM_BY_FILE.INF を置く");
            System.Diagnostics.Process.Start(LOCAL_DIR);

        }

        /// <summary>
        /// 曲目リスト選択
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void listViewMp3_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (NowPlaying == false) {
                txtPlayTime.Text = "";
            }
        }

    }
}
